unit UTree;

interface

uses
  UStack, SysUtils, ComCtrls;

type
  PNode = ^TNode;

  TTree = class
  private
    Root: PNode;
  protected
    procedure DestroyNode; //    
    constructor Create(ch: char); overload; //  -  
    constructor Create(ch: char; Aleft, Aright: TTree); overload; //  - 
    function IsLeaf: boolean; // ,    
    function GetValue: char; //    root
  public
    constructor Create; overload; //   
    destructor Destroy; override; //   
    procedure MakeFromStr(s: string); //     
    procedure MakeForView(Items: TTreeNodes); //    TreeView
    procedure ProcessTree; //  ( )
  end;

  TNode = record
    inf: char;
    left, right: TTree;
  end;

implementation

const
  variables: set of char = ['a' .. 'z', 'A' .. 'Z'];
  signs: set of char = ['+', '-', '*', '/'];
  digits: set of char = ['0' .. '9'];
constructor TTree.Create;    //   
begin
  Root := nil;
end;

constructor TTree.Create(ch: char);  //  -  
begin
  new(Root);
  Root.inf := ch;
  Root.left := nil;
  Root.right := nil;
end;

constructor TTree.Create(ch: char; Aleft, Aright: TTree);  //  - 
begin
  new(Root);
  Root.inf := ch;
  Root.left := Aleft;
  Root.right := Aright;
end;

destructor TTree.Destroy; //   
begin
  if Root <> nil then
  begin
    if Root.left <> nil then
      Root.left.Destroy;
    if Root.right <> nil then
      Root.right.Destroy;
    dispose(Root);
  end;
  inherited Destroy;
end;

procedure TTree.DestroyNode; //    
begin
  if Root <> nil then
  begin
    Root.left := nil;
    Root.right := nil;
  end;
  Destroy;
end;

function TTree.GetValue: char; //    root
begin
  result := Root.inf;
end;

function TTree.IsLeaf: boolean; // ,    
begin
  result := false;
  if Root.left = nil then
    if Root.right = nil then
      result := true;

end;

procedure TTree.MakeForView(Items: TTreeNodes);   //    TreeView

  procedure WriteNodes(var Root: PNode; Node: TTreeNode);
  var                     //     
    Parent: TTreeNode;
  begin
    if Root.inf in signs then
    begin
      Parent := Items.AddChild(Node, Root.inf);
      WriteNodes(Root.left.Root, Parent);
      WriteNodes(Root.right.Root, Parent);
    end
    else
    begin
      Items.AddChild(Node, Root.inf);
    end;
  end;

begin
  Items.Clear;          // 
  WriteNodes(Root, nil);
end;

procedure TTree.MakeFromStr(s: string); //    
var
  i, len: integer;
  St: TStack;
  left, right, Node: TTree;

  procedure SkipSpaces(var s: string; var i: integer); //  
  var
    len: integer;
  begin
    len := length(s);
    while (i <= len) and (s[i] = ' ') do  inc(i);
  end;

begin
  s := Trim(s);
  len := length(s);
  St := TStack.Create;
  i := 1;
  while i <= len do
  begin
    SkipSpaces(s, i);          //   ,   
    if s[i] in signs then      //   ,
    begin
      right := St.Pop;         //   
      left := St.Pop;
      Node := TTree.Create(s[i], left, right);
      St.Push(Node);
    end
    else if (s[i] in (digits + variables)) then
    begin
      Node := TTree.Create(s[i]);  //    
      St.Push(Node);
    end;
    inc(i);
  end;
  Node := St.Pop;      //     
  Root := Node.Root;     //     
  Node.Root := nil;
  Node.DestroyNode; //   
end;

procedure TTree.ProcessTree;  // 
var z:TTree;
begin
  if Root.left <> nil then     //-------
    Root.left.ProcessTree;   //
  if Root.right <> nil then  //
    Root.right.ProcessTree;   //------
  if (getvalue='*') and (not isleaf) then                        //  -----------
    if (root.left.getvalue='+') or (root.left.getvalue='-') then //   
     if not root.left.isleaf then                                //  -----
      begin
       z:=TTree.Create('*',root.left.root.right,root.right); //   -
       root.left.root.right:=root.right;    //  -
       root.right:=z;             // - = *
       root.inf:=root.left.getvalue;  //    ()
       root.left.root.inf:='*';       //   -  
      end;
end;

end.
